from __future__ import annotations

from pathlib import Path

from ..chunk import T
from .types import ChunkPath, Footer, Header


class ChunkStreamReader:
    """A class for reading chunk streams from disk."""

    def __init__(self, manifest_path: str | Path, chunk_type: type[T]) -> None:
        self.manifest_path = Path(manifest_path)
        self.chunk_type = chunk_type
        header, chunk_paths, footer = self.parse_manifest()
        self.header = header
        self.chunk_paths = chunk_paths
        self.footer = footer

    def parse_manifest(self) -> tuple[Header, list[ChunkPath], Footer]:
        """Write the manifest with all the chunks."""
        with open(self.manifest_path) as f:
            value_strings = f.read().split("\n")
        if len(value_strings) < 2:
            raise ValueError(f"Invalid manifest at {self.manifest_path}")
        header = Header.from_json(value_strings[0])
        footer = Footer.from_json(value_strings[-1])
        chunk_paths = list(map(ChunkPath.from_json, value_strings[1:-1]))
        return header, chunk_paths, footer

    def __len__(self) -> int:
        """Get length of chunk stream."""
        return len(self.chunk_paths)

    def __getitem__(self, idx: int) -> T:
        """Get a chunk from a stream."""
        chunk_path = self.manifest_path.parent / self.chunk_paths[idx].path
        return self.chunk_type.from_file(chunk_path)



"""AugmentedBoxTracking chunk."""
from __future__ import annotations

import json
from operator import itemgetter
from pathlib import Path
from typing import Any, Iterator, Optional, Union

import numpy as np
import numpy.typing as npt

from ..boxes import AugmentedBoxes, FrameAugmentedBoxes
from ..types import PType
from .box_tracking import BoxTracking
from .tracking import Tracking


class AugmentedBoxTracking(Tracking, AugmentedBoxes):
    """
    Dense data arrays for box tracking with [boxes, scores, labels].

    It inherits from :class:`.AugmentedBoxes`.
    The underlying NumPy array should have shape ``(n_frames, n_objects, 6)``.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image
    fps : float, optional
        The framerate of the chunk data (if tracking)
    object_ids:  List[str], optional
        Identifiers for the objects (if tracking)
    """

    def validate(self) -> None:
        """Validate tracking shape and boxes."""
        super().validate()
        AugmentedBoxes.validate(self)

    def __iter__(self) -> Iterator[FrameAugmentedBoxes]:
        """Yield FrameAugmentedBoxes objects for each frame."""
        for box in self.view(AugmentedBoxes):
            yield box.view(FrameAugmentedBoxes)

    def to_darwin_dict(
        self,
        filename: str,
        path: str = "/",
        label_names: Optional[list[str]] = None,
    ) -> dict[str, Any]:
        """Serialize boxes to a Darwin annotation dict."""
        if label_names is None:
            label_names = ["Unknown"] * (self.labels.max() + 1)

        annotations: list[dict[str, Any]] = [
            {"frames": {}} for _ in range(self.shape[1])
        ]
        for i, frame in enumerate(self.to_absolute()):
            for j, box in enumerate(frame):
                if np.isnan(box.array).any():
                    continue
                name = box.names(label_names)
                frame_dict = {
                    "bounding_box": {
                        "x": float(box.x),
                        "y": float(box.y),
                        "w": float(box.w),
                        "h": float(box.h),
                    },
                    "keyframe": True,
                }
                annotations[j]["name"] = name
                annotations[j]["frames"][str(i)] = frame_dict
                annotations[j]["id"] = self.object_ids[j]
        for j in range(len(annotations)):
            frame_ids = list(map(int, annotations[j]["frames"].keys()))
            min_frame = min(frame_ids)
            max_frame = max(frame_ids)
            annotations[j]["segments"] = [[min_frame, max_frame]]
        return {
            "image": {
                "filename": filename,
                "path": path,
                "fps": self.fps,
                "width": self.image_width,
                "height": self.image_height,
            },
            "annotations": annotations,
        }

    def to_darwin_file(
        self,
        output_path: Union[str, Path],
        filename: str,
        path: str = "/",
        label_names: Optional[list[str]] = None,
    ) -> None:
        """Write Darwin annotation dict to disk."""
        with open(output_path, "w") as f:
            f.write(
                json.dumps(
                    self.to_darwin_dict(filename, path=path, label_names=label_names)
                )
            )

    @classmethod
    def from_darwin_dict(
        cls: type["AugmentedBoxTracking"],
        darwin_dict: dict[str, Any],
        label_names: list[str] = [],
    ) -> "AugmentedBoxTracking":
        """Create AugmentedBoxTracking chunk from a serialized Darwin dict."""
        label_names_map = {name: float(idx) for idx, name in enumerate(label_names)}
        image_width = darwin_dict["image"]["width"]
        image_height = darwin_dict["image"]["height"]
        fps = darwin_dict["image"]["fps"]
        n_frames = darwin_dict["image"]["frame_count"]
        n_objects = len(darwin_dict["annotations"])
        data: npt.NDArray[np.float64] = np.zeros((n_frames, n_objects, 6)) * np.nan
        score = 1.0
        object_ids: list[str] = [t["id"] for t in darwin_dict["annotations"]]
        for i in range(n_frames):
            for j, tracklet in enumerate(darwin_dict["annotations"]):
                annotation = tracklet["frames"].get(str(i), {})
                if "bounding_box" not in annotation:
                    continue
                x, y, w, h = itemgetter("x", "y", "w", "h")(annotation["bounding_box"])
                label = label_names_map.get(tracklet["name"], -1.0)
                data[i, j] = x, y, w, h, score, label
        return cls(
            data,
            ptype=PType.absolute_tlwh,
            image_width=image_width,
            image_height=image_height,
            fps=fps,
            object_ids=object_ids,
        )

    @classmethod
    def from_darwin_file(
        cls: type["AugmentedBoxTracking"],
        path: Union[str, Path],
        label_names: list[str] = [],
    ) -> "AugmentedBoxTracking":
        """Read AugmentedBoxTracking from Darwin dict on disk."""
        with open(path) as f:
            darwin_dict = json.loads(f.read())
        return cls.from_darwin_dict(darwin_dict, label_names=label_names)

    @classmethod
    def from_dict(
        cls: type["AugmentedBoxTracking"],
        chunk_dict: dict[str, Any],
        dims: Optional[int] = None,
    ) -> "AugmentedBoxTracking":
        """Create chunk from chunk dict."""
        return super().from_dict(chunk_dict, dims=6)

    @classmethod
    def from_frame_augmented_boxes(
        cls: type["AugmentedBoxTracking"],
        boxes: list[FrameAugmentedBoxes],
        ptype: PType = PType.unknown,
        image_width: Optional[int] = None,
        image_height: Optional[int] = None,
        fps: Optional[float] = None,
        object_ids: Optional[list[str]] = None,
    ) -> "AugmentedBoxTracking":
        """
        Create an AugmentedBoxTracking chunk from a list of FrameAugmentedBoxes objects.

        This is typically used for storing detections on a video.
        """
        max_boxes = max(map(len, boxes))
        data = np.zeros((len(boxes), max_boxes, 6)) * np.nan
        for i, frame in enumerate(boxes):
            data[i, : len(frame)] = frame.array
        return cls(
            data,
            ptype=ptype,
            image_width=image_width,
            image_height=image_height,
            fps=fps,
            object_ids=object_ids,
        )

    def filter_by_class(self, class_idx: int) -> BoxTracking:
        """Get a single class and return a new BoxTracking chunk."""
        object_mask: list[int] = []
        for j in range(self.shape[1]):
            if class_idx in set(self.array[:, j, -1].ravel()):
                object_mask.append(j)
        return BoxTracking(
            self.array[:, np.array(object_mask), :4],
            ptype=self.ptype,
            **self.metadata_kwargs,
        )

    @classmethod
    def from_box_tracking(
        cls: type["AugmentedBoxTracking"],
        chunk: BoxTracking,
        score: float = 0.0,
        class_idx: int = 0,
    ) -> "AugmentedBoxTracking":
        """Create AugmentedBoxTracking chunk from BoxTracking chunk."""
        pad_shape = np.ones(chunk.shape[:2] + (1,))
        data = np.concatenate(
            [
                chunk.array[..., :4],
                pad_shape * score,
                pad_shape * class_idx,
            ],
            axis=-1,
        )
        return cls(data, ptype=chunk.ptype, **chunk.metadata_kwargs)



"""Box chunks."""
from .augmented_boxes import AugmentedBoxes
from .boxes import Boxes
from .frame_augmented_boxes import FrameAugmentedBoxes
from .frame_boxes import FrameBoxes
from .iou import area, intersection, pairwise_iou
from .nms import non_max_suppression
from .single_augmented_box import SingleAugmentedBox
from .single_box import SingleBox



from .augmented_boxes import AugmentedBoxes
from .single_box import SingleBox, _is_1d


class SingleAugmentedBox(AugmentedBoxes):
    """
    1D dense data arrays for an augmented box.

    The data contain ``[box, score, label]`` components.
    It inherits from :class:`.AugmentedBoxes`.
    The underlying NumPy array should have shape ``(6,)``.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image
    """

    def validate(self) -> None:
        """Check validity of boxes array."""
        super().validate()
        _is_1d(self)

    @property
    def score(self) -> float:
        """Singular alias for scores."""
        return float(self.scores)

    @property
    def label(self) -> int:
        """Singular alias for labels."""
        return int(self.labels)

    def to_single_box(self) -> SingleBox:
        """Convert to SingleBox."""
        return SingleBox(
            self.array[:4],
            ptype=self.ptype,
            **self.metadata_kwargs,
        )



"""Non-max suppression."""
import numpy as np

from .frame_augmented_boxes import FrameAugmentedBoxes


def non_max_suppression(
    boxes: FrameAugmentedBoxes,
    minimum_score: float = 0.5,
    iou_threshold: float = 0.5,
) -> FrameAugmentedBoxes:
    """Suppress overlapping boxes by box confidence and IoU."""
    keepers: list[int] = []
    score_mask = boxes.scores > minimum_score
    class_indices = sorted(set(boxes.labels.ravel()))
    for class_index in class_indices:
        mask = np.logical_and(score_mask, boxes.labels == class_index)
        mask_indices = np.argwhere(mask).ravel()
        box_candidates = boxes[mask]
        areas = (box_candidates.w + 1) * (box_candidates.h + 1)
        score_order = box_candidates.scores.argsort()[::-1]
        while score_order.size > 0:
            i = score_order[0]
            keepers.append(mask_indices[i])
            xx1 = np.maximum(box_candidates.x1[i], box_candidates.x1[score_order[1:]])
            yy1 = np.maximum(box_candidates.y1[i], box_candidates.y1[score_order[1:]])
            xx2 = np.minimum(box_candidates.x2[i], box_candidates.x2[score_order[1:]])
            yy2 = np.minimum(box_candidates.y2[i], box_candidates.y2[score_order[1:]])

            w = np.maximum(0.0, xx2 - xx1 + 1)  # maximum width
            h = np.maximum(0.0, yy2 - yy1 + 1)  # maxiumum height
            intersection = w * h
            iou = intersection / (areas[i] + areas[score_order[1:]] - intersection)

            index = np.where(iou <= iou_threshold)[0]
            score_order = score_order[index + 1]

    if len(keepers):
        return boxes[np.array(keepers)]
    return FrameAugmentedBoxes(
        np.zeros((0, 6)),
        ptype=boxes.ptype,
        **boxes.metadata_kwargs,
    )



from typing import TypeVar

import numpy as np

from ..chunk import Chunk
from ..types import FloatArray, PType

T = TypeVar("T", bound="Boxes")


class Boxes(Chunk):
    """
    Dense data arrays for boxes.

    This mostly serves as a base class for more specific types of box chunks.
    It inherits from :class:`.Chunk`.
    The underlying NumPy array should have shape ``(..., 4)``.

    Parameters
    ----------
    data : FloatArray
        A numpy array of dense floats
    ptype : PType
        The parameterization of the dense data
    image_width : int, optional
        The width of the relevant image
    image_height : int, optional
        The height of the relevant image

    Example
    -------
    >>> import numpy as np
    >>> from sparrow_datums import Boxes, PType
    >>> Boxes(np.ones(4), PType.absolute_tlbr).to_tlwh()
    Boxes([1., 1., 0., 0.])
    """

    def validate(self) -> None:
        """Check validity of boxes array."""
        if not self.shape or self.shape[-1] != 4:
            raise ValueError("Box arrays must have size-4 dimensions")

    @property
    def is_relative(self) -> bool:
        """Parameterization is relative."""
        return bool(self.ptype.is_relative)

    @property
    def is_absolute(self) -> bool:
        """Parameterization is absolute."""
        return bool(self.ptype.is_absolute)

    @property
    def is_tlbr(self) -> bool:
        """Parameterization is TLBR."""
        return bool(self.ptype.is_tlbr)

    @property
    def is_tlwh(self) -> bool:
        """Parameterization is TLWH."""
        return bool(self.ptype.is_tlwh)

    def box_resize(self: T, image_width: int, image_height: int) -> T:
        """Resize boxes to a new width and height."""
        relative_boxes = self.to_relative()
        new_scale = np.array([image_width, image_height, image_width, image_height])
        x = relative_boxes.array.copy()
        x[..., :4] *= new_scale
        metadata_kwargs = self.metadata_kwargs.copy()
        metadata_kwargs["image_width"] = image_width
        metadata_kwargs["image_height"] = image_height
        return self.__class__(
            x,
            ptype=self.ptype.as_absolute,
            **metadata_kwargs,
        )

    def to_relative(self: T) -> T:
        """Convert boxes to relative pixel coordinates, if necessary."""
        if self.is_relative:
            return self
        x = self.array.copy()
        x[..., :4] /= self.scale
        return self.__class__(
            x,
            ptype=self.ptype.as_relative,
            **self.metadata_kwargs,
        )

    def to_absolute(self: T) -> T:
        """Convert boxes to absolute pixel coordinates, if necessary."""
        if self.is_absolute:
            return self
        x = self.array.copy()
        x[..., :4] *= self.scale
        return self.__class__(
            x,
            ptype=self.ptype.as_absolute,
            **self.metadata_kwargs,
        )

    def to_tlbr(self: T) -> T:
        """Convert boxes to TLBR format, if necessary."""
        if self.is_tlbr:
            return self
        x = self.array[..., 0]
        y = self.array[..., 1]
        w = self.array[..., 2]
        h = self.array[..., 3]
        return self.__class__(
            np.concatenate(
                [np.stack([x, y, x + w, y + h], -1), self.array[..., 4:]], -1
            ),
            ptype=self.ptype.as_tlbr,
            **self.metadata_kwargs,
        )

    def to_tlwh(self: T) -> T:
        """Convert boxes to TLWH format, if necessary."""
        if self.is_tlwh:
            return self
        x1 = self.array[..., 0]
        y1 = self.array[..., 1]
        x2 = self.array[..., 2]
        y2 = self.array[..., 3]
        return self.__class__(
            np.concatenate(
                [np.stack([x1, y1, x2 - x1, y2 - y1], -1), self.array[..., 4:]], -1
            ),
            ptype=self.ptype.as_tlwh,
            **self.metadata_kwargs,
        )

    def validate_known_ptype(self) -> None:
        """Make sure PType is a known box parameterization."""
        known_box_parameterizations = {
            PType.absolute_tlbr,
            PType.absolute_tlwh,
            PType.relative_tlbr,
            PType.relative_tlwh,
        }
        if self.ptype not in known_box_parameterizations:
            raise ValueError(f"Unknown box parameterization: {self.ptype.name}")

    @property
    def x(self) -> FloatArray:
        """Slice the x dimension of the boxes."""
        self.validate_known_ptype()
        result: FloatArray
        result = self.array[..., 0]
        return result

    @property
    def y(self) -> FloatArray:
        """Slice the y dimension of the boxes."""
        self.validate_known_ptype()
        result: FloatArray
        result = self.array[..., 1]
        return result

    @property
    def w(self) -> FloatArray:
        """Slice the width dimension of the boxes."""
        self.validate_known_ptype()
        result: FloatArray
        if self.is_tlwh:
            result = self.array[..., 2]
            return result
        result = self.array[..., 2] - self.array[..., 0]
        return result

    @property
    def h(self) -> FloatArray:
        """Slice the height dimension of the boxes."""
        self.validate_known_ptype()
        result: FloatArray
        if self.is_tlwh:
            result = self.array[..., 3]
            return result
        result = self.array[..., 3] - self.array[..., 1]
        return result

    @property
    def x1(self) -> FloatArray:
        """Slice the x1 dimension of the boxes."""
        self.validate_known_ptype()
        return self.x

    @property
    def y1(self) -> FloatArray:
        """Slice the y1 dimension of the boxes."""
        self.validate_known_ptype()
        return self.y

    @property
    def x2(self) -> FloatArray:
        """Slice the x2 dimension of the boxes."""
        self.validate_known_ptype()
        result: FloatArray
        if self.is_tlbr:
            result = self.array[..., 2]
            return result
        result = self.array[..., 0] + self.array[..., 2]
        return result

    @property
    def y2(self) -> FloatArray:
        """Slice the y2 dimension of the boxes."""
        self.validate_known_ptype()
        result: FloatArray
        if self.is_tlbr:
            result = self.array[..., 3]
            return result
        result = self.array[..., 1] + self.array[..., 3]
        return result



import doctest

import numpy as np
import pytest

from ..types import FloatArray, PType
from . import boxes
from .boxes import Boxes


def test_docstring_example():
    result = doctest.testmod(boxes)
    assert result.attempted > 0
    assert result.failed == 0


def test_bad_shape_throws_type_error():
    with pytest.raises(ValueError):
        Boxes(np.ones(3))


def test_to_relative_moves_boxes_to_0_1():
    image_size = 512
    boxes_a = Boxes(
        np.random.uniform(size=(10, 4)) * image_size,
        PType.absolute_tlwh,
        image_width=image_size,
        image_height=image_size,
    )
    boxes_b = boxes_a.to_relative()
    assert boxes_b.is_relative
    np.testing.assert_array_less(boxes_b.array, 1.0)
    boxes_c = boxes_b.to_relative()
    assert boxes_c.is_relative


def test_to_absolute_moves_boxes_to_0_image_size():
    image_size = 512
    boxes_a = Boxes(
        np.random.uniform(size=(10, 4)),
        PType.relative_tlwh,
        image_width=image_size,
        image_height=image_size,
    )
    boxes_b = boxes_a.to_absolute()
    assert boxes_b.is_absolute
    assert (boxes_b.array < 1).mean() < 0.1
    boxes_c = boxes_b.to_absolute()
    assert boxes_c.is_absolute


def test_to_tlbr_moves_boxes_to_tlbr():
    x: FloatArray = np.concatenate([np.ones((5, 2)), np.zeros((5, 2))], -1)
    boxes = Boxes(x, PType.relative_tlwh).to_tlbr()
    np.testing.assert_equal(boxes.array, 1)


def test_to_tlwh_moves_boxes_to_tlwh():
    result: FloatArray = np.concatenate([np.ones((5, 2)), np.zeros((5, 2))], -1)
    boxes = Boxes(np.ones((5, 4)), PType.relative_tlbr).to_tlwh()
    np.testing.assert_equal(boxes.array, result)


def test_boxes_to_tlbr_doesnt_change_original_instance():
    x: FloatArray = np.concatenate([np.ones((5, 2)), np.zeros((5, 2))], -1)
    boxes = Boxes(x, PType.relative_tlwh)
    _ = boxes.to_tlbr()
    assert boxes.is_tlwh
    assert boxes.is_relative


def test_box_slicing_cols_throws_value_error():
    boxes = Boxes(np.random.uniform(size=(10, 4)))
    with pytest.raises(ValueError):
        boxes[:, :2]
    top_left = boxes.array[:, :2]
    assert isinstance(top_left, np.ndarray)


def test_boxes_deserializes_type():
    boxes = Boxes.from_dict(
        {
            "data": [1, 2, 3, 4],
            "ptype": "relative_tlbr",
            "image_width": None,
            "image_height": None,
            "fps": None,
            "object_ids": None,
        }
    )
    assert boxes.ptype == PType.relative_tlbr


def test_boxes_saves_classname():
    data = Boxes(np.ones(4)).to_dict()
    assert data["classname"] == "Boxes"


def test_box_attributes_require_known_box_parameterization():
    with pytest.raises(ValueError):
        Boxes(np.ones(4)).x


def test_resize():
    boxes = Boxes(np.ones(4), ptype=PType.absolute_tlwh, image_width=4, image_height=4)
    new_boxes = boxes.box_resize(32, 16)
    assert new_boxes.ptype == PType.absolute_tlwh
    assert new_boxes.image_width == 32
    assert new_boxes.image_height == 16
    np.testing.assert_equal(new_boxes.array, np.array([8, 4, 8, 4]))



import doctest

import numpy as np
import pytest

from ..types import FloatArray, PType
from . import frame_boxes
from .frame_boxes import FrameBoxes
from .single_box import SingleBox


def test_docstring_example():
    result = doctest.testmod(frame_boxes)
    assert result.attempted > 0
    assert result.failed == 0


def test_frame_boxes_conversion_creates_frame_boxes():
    boxes_a = FrameBoxes(np.ones((2, 4)), PType.relative_tlbr)
    boxes_b = boxes_a.to_tlwh()
    assert isinstance(boxes_b, FrameBoxes)


def test_frame_boxes_with_bad_shape_throws_value_error():
    with pytest.raises(ValueError):
        FrameBoxes(np.ones(4))
    with pytest.raises(ValueError):
        FrameBoxes(np.ones((2, 3)))


def test_frame_boxes_iterator_makes_single_boxes():
    boxes = FrameBoxes(np.ones((2, 4)), PType.relative_tlbr)
    for box in boxes:
        assert isinstance(box, SingleBox)
        assert box.ptype == PType.relative_tlbr


def test_from_single_box_preserves_data():
    box = SingleBox(np.random.uniform(size=4))
    frame_boxes = FrameBoxes.from_single_box(box)
    np.testing.assert_equal(box.array, frame_boxes.array.ravel())
    assert box.ptype == frame_boxes.ptype


def test_add_box_with_different_attributes_fails():
    data: FloatArray = np.random.uniform(size=4)
    boxes = FrameBoxes.from_single_box(SingleBox(data))
    box_b = SingleBox(data, ptype=PType.relative_tlbr)
    with pytest.raises(ValueError):
        boxes.add_box(box_b)
    box_c = SingleBox(data, image_width=100)
    with pytest.raises(ValueError):
        boxes.add_box(box_c)


def test_add_box_with_same_attributes_works():
    data: FloatArray = np.random.uniform(size=4)
    boxes = FrameBoxes.from_single_boxes([SingleBox(data)])
    box_b = SingleBox(data + 1)
    new_boxes = boxes.add_box(box_b)
    assert len(new_boxes) == 2


def test_get_single_box_returns_single_box():
    boxes = FrameBoxes(np.ones((2, 4)), PType.relative_tlbr)
    single_box = boxes.get_single_box(0)
    assert isinstance(single_box, SingleBox)
    assert single_box.ptype == boxes.ptype



import numpy as np
import pytest

from .tracking import Tracking


def test_validate_tracking_chunk():
    with pytest.raises(ValueError):
        Tracking(np.ones((2, 4)))
    Tracking(np.ones((2, 2, 2)))


def test_resample_random_chunk():
    x = np.random.uniform(size=(10, 5, 3))
    chunk = Tracking(x, fps=20)
    fast_chunk = chunk.resample(40)
    assert fast_chunk.fps == 40
    assert len(fast_chunk) == 20
    slow_chunk = fast_chunk.resample(20)
    np.testing.assert_equal(chunk.array, slow_chunk.array)


def test_valid_duration():
    tracking = Tracking(np.ones((2, 2, 2)), fps=30)
    assert tracking.duration == 1 / 15
